#!/usr/bin/env python
# encoding: utf-8

"""Utilities."""

from ConfigParser import (NoOptionError, NoSectionError, ParsingError,
  RawConfigParser)
from contextlib import contextmanager
from functools import wraps
from logging.handlers import TimedRotatingFileHandler
from os import close, remove
from shutil import rmtree
from tempfile import gettempdir, mkstemp
import logging as lg
import os.path as osp
import sys
import warnings as wr


_logger = lg.getLogger(__name__)


class HdfsError(Exception):

  """Base error class.

  :param message: Error message.
  :param args: optional Message formatting arguments.

  """

  def __init__(self, message, *args):
    super(HdfsError, self).__init__(message % args if args else message)


class InstanceLogger(lg.LoggerAdapter):

  """Logger adapter that preprends the instance's `repr` to messages.

  :param instance: Instance to prepend messages with.
  :param logger: Logger instance where messages will be logged.
  :param extra: Dictionary of contextual information, passed to the formatter.

  """

  def __init__(self, instance, logger, extra=None):
    # not using super since `LoggerAdapter` is an old-style class in python 2.6
    lg.LoggerAdapter.__init__(self, logger, extra)
    self.instance = instance

  def process(self, msg, kwargs):
    """Transform message.

    :param msg: Original message.
    :param kwargs: Dictionary of kwargs eventually passed to the formatter.

    """
    return '%r :: %s' % (self.instance, msg), kwargs


class Config(object):

  """Configuration class.

  :param path: path to configuration file. If no file exists at that location,
    the configuration parser will be empty.

  """

  def __init__(self, path=osp.expanduser('~/.hdfsrc')):
    self._logger = InstanceLogger(self, _logger)
    self.parser = RawConfigParser()
    self.path = path
    if osp.exists(path):
      try:
        self.parser.read(self.path)
      except ParsingError:
        raise HdfsError('Invalid configuration file %r.', path)

  def __repr__(self):
    return '<Config(path=%r)>' % (self.path, )

  def save(self):
    """Save configuration parser back to file."""
    with open(self.path, 'w') as writer:
      self.parser.write(writer)
    self._logger.info('Saved.')

  def get_alias(self, alias=None):
    """Retrieve alias information from configuration file.

    :param alias: Alias name. If not specified, will use the `default.alias` in
      the `hdfs` section if provided, else will raise :class:`HdfsError`.

    Raises :class:`HdfsError` if no matching / an invalid alias was found.

    """
    try:
      alias = alias or self.parser.get('hdfs', 'default.alias')
    except (NoOptionError, NoSectionError):
      raise HdfsError('No alias specified and no default alias found.')
    section = '%s_alias' % (alias, )
    try:
      options = dict(self.parser.items(section))
    except NoSectionError:
      raise HdfsError('Alias not found: %r.', alias)
    else:
      return options

  def get_file_handler(self, name):
    """Create and configure logging file handler.

    :param name: Section name used to find the path to the log file. If no
      `log` option exists in this section, the path will default to
      `<name>.log`.

    The default path can be configured via the `default.log` option in the
    `hdfs` section.

    """
    try:
      handler_path = self.parser.get(name, 'log')
    except (NoOptionError, NoSectionError):
      handler_path = osp.join(gettempdir(), '%s.log' % (name, ))
    try:
      handler = TimedRotatingFileHandler(
        handler_path,
        when='midnight', # daily backups
        backupCount=1,
        encoding='utf-8',
      )
    except IOError:
      wr.warn('Unable to write to log file at %s.' % (handler_path, ))
    else:
      handler_format = (
        '%(asctime)s | %(levelname)4.4s | %(name)s > %(message)s'
      )
      handler.setFormatter(lg.Formatter(handler_format))
      return handler


@contextmanager
def temppath(dir=None):
  """Create a temporary path.

  :param dir: Explicit directory name where to create the temporary path. A
    system dependent default will be used otherwise (cf. `tempfile.mkstemp`).

  Usage::

    with temppath() as path:
      pass # do stuff

  Any file or directory corresponding to the path will be automatically deleted
  afterwards.

  """
  (desc, path) = mkstemp(dir=dir)
  close(desc)
  remove(path)
  try:
    _logger.debug('Created temporary path at %s.', path)
    yield path
  finally:
    if osp.exists(path):
      if osp.isdir(path):
        rmtree(path)
        _logger.debug('Deleted temporary directory at %s.', path)
      else:
        remove(path)
        _logger.debug('Deleted temporary file at %s.', path)
    else:
      _logger.debug('No temporary file or directory to delete at %s.', path)

def hsize(size):
  """Transform size from bytes to human readable format (kB, MB, ...).

  :param size: Size in bytes.

  """
  for suffix in [' ', 'k', 'M', 'G', 'T']:
    if size < 1024.0:
      return '%4.0f%sB' % (size, suffix)
    size /= 1024.0

def htime(time):
  """Transform time from seconds to human readable format (min, hour, ...).

  :param time: Time in seconds.

  """
  for (multiplier, suffix) in [
    (60.0, 's'),
    (60, 'm'),
    (24, 'h'),
    (7, 'd'),
    (4, 'w'),
    (12, 'M'),
    (float('inf'), 'Y'),
  ]:
    if time < multiplier:
      return '%4.1f%s' % (time, suffix)
    time /= multiplier

def catch(*error_classes):
  """Returns a decorator that catches errors and prints messages to stderr.

  :param \*error_classes: Error classes.

  Also exits with status 1 if any errors are caught.

  """
  def decorator(func):
    """Decorator."""
    @wraps(func)
    def wrapper(*args, **kwargs):
      """Wrapper. Finally."""
      try:
        return func(*args, **kwargs)
      except error_classes as err:
        _logger.error(err)
        sys.stderr.write('%s\n' % (err, ))
        sys.exit(1)
      except Exception as err: # catch all
        _logger.exception('Unexpected exception.')
        sys.stderr.write(
          'Unexpected error: %s\n'
          'View log for more details.\n' % (err, )
        )
        sys.exit(1)
    return wrapper
  return decorator



#!/usr/bin/env python
# encoding: utf-8

"""Test Hdfs Dataframe extension."""

from helpers import _TestSession
import os
import shutil
import tempfile
from hdfs.util import HdfsError
from nose.tools import raises
try:
  from hdfs.ext.dataframe import *
except ImportError:
  SKIP = True
else:
  from pandas.util.testing import assert_frame_equal
  import pandas as pd
  import numpy
  SKIP = False


class TestDataframe(_TestSession):

  def setup(self):
    if SKIP:
      self.client = None
    super(TestDataframe, self).setup()
    self.test_df = pd.DataFrame.from_records(
        [{'A' :  1, 'B' :  2},
        {'A' : 11, 'B' : 23},
        {'A' :100, 'B' :211},
        {'A' : 23, 'B' :  1}])

  def run_write_read(self, df, format, use_gzip = False, sep = '\t', 
      index_cols = None, local_dir = None, n_threads = None,
      hdfs_filename = 'dfreader_test', read_hdfs_filename=None, n_parts = 2):

    # Location on HDFS
    ext = format + ('.gz' if use_gzip else '')

    write_df(df, self.client, hdfs_filename, format, sep=sep, use_gzip=use_gzip, 
      overwrite=True, n_parts=n_parts)

    if read_hdfs_filename is not None:
      r_filename = read_hdfs_filename
    else:
      r_filename = hdfs_filename
    returned_df = read_df(self.client, r_filename, format, sep=sep, 
      use_gzip=use_gzip, index_cols=index_cols, local_dir=local_dir, 
      n_threads=n_threads)
    assert_frame_equal(df, returned_df)
    return returned_df

  def test_csv(self):
    self.run_write_read(self.test_df, format='csv')

  def test_sep_suffix(self):
    self.run_write_read(self.test_df, format='csv', 
      hdfs_filename='dfreader_test/test/')

  def test_dir_with_single_part(self):
    self.run_write_read(self.test_df, format='csv', n_parts=1)

  @raises(HdfsError)
  def test_single_file(self):
    self.run_write_read(self.test_df, format='csv', 
      hdfs_filename='dfreader_test/test',
      read_hdfs_filename='dfreader_test/test/part-r-00000')

  def test_csv_gz(self):
    self.run_write_read(self.test_df, format='csv', use_gzip=True)

  def test_csv_comma(self):
    self.run_write_read(self.test_df, format='csv', sep=',')

  def test_csv_index(self):
    df = self.test_df.copy().set_index('A')
    self.run_write_read(df, format='csv', index_cols = ['A'])

  def test_avro(self):
    self.run_write_read(self.test_df, format='avro')

  def test_avro_index(self):
    df = self.test_df.copy().set_index('A')
    self.run_write_read(df, format='avro', index_cols = ['A'])

  def test_local_tmp_path(self):
    temp1 = tempfile.mkdtemp()
    try:
      returned_df = self.run_write_read(self.test_df, 'csv', temp1)

      temp2 = tempfile.mkdtemp()
      try:
        # Should be cached now
        self.run_write_read(returned_df, 'csv', temp2)

      finally:
        shutil.rmtree(temp2)

    finally:
      shutil.rmtree(temp1)

  def test_nonparallel_download(self):
    self.run_write_read(self.test_df, 'csv', n_threads = 1)

  def test_parallel_download(self):
    self.run_write_read(self.test_df, 'csv', n_threads = -1)




#!/usr/bin/env python
# encoding: utf-8

"""Test HdfsAvro extension."""

from helpers import _TestSession
from nose.tools import *
try:
  from hdfs.ext.avro import *
except ImportError:
  AvroTypeException = Exception
  SKIP = True
else:
  from avro.io import AvroTypeException
  SKIP = False


class TestWriter(_TestSession):

  def setup(self):
    if SKIP:
      self.client = None
    super(TestWriter, self).setup()
    # if we reach here, self.client is defined
    self.writer = AvroWriter(self.client, 'aw.avro')

  def test_write_inferring_schema(self):
    with self.writer as writer:
      self.writer.records.send({'foo': 'value1'})
      self.writer.records.send({'foo': 'value2', 'bar': 'that'})
    data = self.client._open('aw.avro').content
    ok_('foo' in data)
    ok_('value1' in data)
    ok_('value2' in data)
    ok_(not 'that' in data)

  @raises(AvroTypeException)
  def test_invalid_schema(self):
    with self.writer as writer:
      self.writer.records.send({'foo': 'value1'})
      self.writer.records.send({'bar': 'value2'})



#!/usr/bin/env python
# encoding: utf-8

"""HdfsCLI."""

__version__ = '1.0.0'

import logging as lg
try:
  from .client import Client, InsecureClient, TokenClient
  from .ext import * # Import all symbols exported by extensions.
except ImportError:
  pass # In setup.py.


class _NullHandler(lg.Handler):

  """For python <2.7."""

  def emit(self, record):
    pass


lg.getLogger(__name__).addHandler(_NullHandler())



#!/usr/bin/env python
# encoding: utf-8

"""HdfsAvro: an Avro extension for HdfsCLI.

Usage:
  hdfsavro [-a ALIAS] --schema PATH
  hdfsavro [-a ALIAS] --head [-n NUM] [-p PARTS] PATH
  hdfsavro [-a ALIAS] --sample (-f FREQ | -n NUM) PATH
  hdfsavro -h | --help | -l | --log

Commands:
  --schema                      Pretty print schema.
  --head                        Pretty print first few records as JSON.
  --sample                      Sample records and output JSON.

Arguments:
  PATH                          Remote path to Avro file.

Options:
  -a ALIAS --alias=ALIAS        Alias.
  -h --help                     Show this message and exit.
  -f FREQ --freq=FREQ           Probability of sampling a record.
  -l --log                      Show path to current log file and exit.
  -n NUM --num=NUM              Number of records to output [default: 5].
  -p PARTS --parts=PARTS        Part-files. `1,` to get a unique part-file. By
                                default, use all part-files.

"""

from __future__ import absolute_import
from ..client import Client
from ..util import Config, HdfsError, catch
from docopt import docopt
from io import BytesIO
from itertools import islice
from json import dumps
from random import random, shuffle
from tempfile import mkstemp
import avro as av
import avro.datafile as avd
import avro.io as avi
import logging as lg
import os
import sys
import zlib
try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO
try:
  import snappy
except ImportError:
  pass


def _get_type(obj, allow_null=False):
  """Infer Avro type corresponding to a python object.

  :param obj: Python primitive.
  :param allow_null: Allow null values.

  """
  if allow_null:
    raise NotImplementedError('TODO')
  if isinstance(obj, bool):
    schema_type = 'boolean'
  elif isinstance(obj, basestring):
    schema_type = 'string'
  elif isinstance(obj, int):
    schema_type = 'int'
  elif isinstance(obj, long):
    schema_type = 'long'
  elif isinstance(obj, float):
    schema_type = 'float'
  return schema_type


def _get_schema(dct, allow_null=False):
  """Infer schema from dictionary.

  :param dct: Dictionary.
  :param allow_null: Allow null values.

  """
  fields = [
    '{"name": "%s", "type": "%s"}' % (k, _get_type(v, allow_null=allow_null))
    for k, v in dct.items()
  ]
  return av.schema.parse(
    """
      {
        "type": "record",
        "name": "element",
        "fields": [%s]
      }
    """ % (','.join(fields), )
  )


class _LazyBinaryDecoder(avi.BinaryDecoder):

  """Custom BinaryDecoder that can read lazily from a remote source.

  :param reader: Output of :meth:`hdfs.client.Client.read`.
  :param path: Path to file used as local buffer.

  """

  reader = None # override read-only property

  def __init__(self, reader, path):
    self.file = open(path, 'w+')
    self.reader = reader
    self.loaded = 0
    self.exhausted = False

  def _load(self, size):
    """Load more data. Should only be called when at the end of the file!"""
    pos = self.file.tell()
    while size > 0:
      try:
        chunk = self.reader.next()
      except StopIteration:
        self.exhausted = True
        break
      else:
        size -= len(chunk)
        self.file.write(chunk)
    self.file.seek(pos)

  def read(self, size):
    """Read from file. We use the fact that size is always specified here."""
    data = self.file.read(size)
    size -= len(data)
    if size and not self.exhausted:
      self._load(size)
      data += self.file.read(size)
    return data

  def seek(self, offset, whence=None):
    """Jump to position in file."""
    self.file.seek(offset, whence)

  def skip(self, size):
    """Skip bytes."""
    missing = len(self.file.read(size)) - size
    if missing and not self.exhausted:
      self._load(missing)
      self.file.seek(missing)


class _DatumReader(avi.DatumReader):

  """DatumReader that HEX-encodes "fixed" fields.

  If we don't do this, the JSON encoder will blow up when faced with non-
  unicode characters. This is simpler and more efficient than the alternative
  of writing our custom JSON encoder.

  """

  def read_fixed(self, writers_schema, readers_schema, decoder):
    """Return HEX-encoded value instead of raw byte-string."""
    return '0x' + decoder.read(writers_schema.size).encode('hex')


class _DataFileReader(object):

  """Read remote avro file.

  :param reader: Output of :meth:`~hdfs.client.Client.read`.

  This is a more efficient implementation of the :class:`~avro.DataFileReader`,
  customized for reading remote files. The default one reads the entire file
  before emitting any records which is dramatically slow in our case.

  It should be used as a context manager. E.g.

  .. code-block:: python

    reader = client.read('foo.avro')
    with _DataFileReader(reader) as avro_reader:
      schema = avro_reader.schema
      for record in avro_reader:
        pass # do something ...

  """

  def __init__(self, reader):
    self.schema = None
    self._reader = reader
    self._datum_reader = _DatumReader()
    self._raw_decoder = None
    self._datum_decoder = None
    self._block_count = 0
    self._sync_marker = None
    self._meta = None
    self._codec = None
    self._path = None

  def __enter__(self):
    # create temporary file to serve as buffer
    desc, self._path = mkstemp()
    os.close(desc)
    self._raw_decoder = _LazyBinaryDecoder(self._reader, self._path)
    # load header data
    header = self._read_header()
    self._sync_marker = header['sync'][2:].decode('hex')
    self._meta = header['meta']
    self._codec = self._meta['avro.codec']
    # get ready to read by attaching schema to the datum reader
    self.schema = av.schema.parse(self._meta.get(avd.SCHEMA_KEY))
    self._datum_reader.writers_schema = self.schema
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    self._reader.close()
    os.remove(self._path)

  def __iter__(self):
    return self

  def _read_header(self):
    """Parse header into dictionary."""
    header = self._datum_reader.read_data(
      avd.META_SCHEMA, avd.META_SCHEMA, self._raw_decoder
    )
    magic = header.get('magic')[2:].decode('hex')
    if magic != avd.MAGIC:
      msg = 'Invalid magic #: %s doesn\'t match %s.' % (magic, avd.MAGIC)
      raise av.schema.AvroException(msg)
    codec = header['meta'].setdefault('avro.codec', 'null')
    if codec not in avd.VALID_CODECS:
      raise avd.DataFileException('Unknown codec: %r.' % (codec, ))
    return header

  def _read_block_header(self):
    """Find block count and set the decoder accordingly."""
    if self._codec == 'null':
      # Skip a long; we don't need to use the length.
      self._raw_decoder.skip_long()
      self._datum_decoder = self._raw_decoder
    elif self._codec == 'deflate':
      # Compressed data is stored as (length, data), which
      # corresponds to how the "bytes" type is encoded.
      data = self._raw_decoder.read_bytes()
      # -15 is the log of the window size; negative indicates
      # "raw" (no zlib headers) decompression.  See zlib.h.
      uncompressed = zlib.decompress(data, -15)
      self._datum_decoder = avi.BinaryDecoder(StringIO(uncompressed))
    elif self._codec == 'snappy':
      # Compressed data includes a 4-byte CRC32 checksum
      length = self._raw_decoder.read_long()
      data = self._raw_decoder.read(length - 4)
      uncompressed = snappy.decompress(data)
      self._datum_decoder = avi.BinaryDecoder(StringIO(uncompressed))
      self._raw_decoder.check_crc32(uncompressed)
    else:
      raise avd.DataFileException("Unknown codec: %r" % self._codec)

  def next(self):
    """Return the next datum in the file."""
    if not self._block_count:
      proposed_sync_marker = self._raw_decoder.read(avd.SYNC_SIZE)
      if proposed_sync_marker != self._sync_marker:
        self._raw_decoder.seek(-avd.SYNC_SIZE, 1)
      try:
        self._block_count = self._raw_decoder.read_long()
      except TypeError: # end of file
        raise StopIteration
      else:
        self._read_block_header()
    datum = self._datum_reader.read(self._datum_decoder)
    self._block_count -= 1
    return datum


class AvroReader(object):

  """Lazy remote Avro file reader.

  :param client: :class:`hdfs.client.Client` instance.
  :param hdfs_path: Remote path.
  :param parts: Cf. :meth:`hdfs.client.Client.parts`.

  Usage:

  .. code-block:: python

    with AvroReader(client, 'foo.avro') as reader:
      schema = reader.schema # the remote file's Avro schema
      for record in reader.records:
        pass # and its records

  """

  records = None
  """Avro record generator.

  This generator will yield a dictionary for each record in the remote Avro
  file. For convenience, you can also iterate directly on the
  :class:`AvroReader` object. E.g.

  .. code-block:: python

    with AvroReader(client, 'foo.avro') as reader:
      for record in reader:
        print record.to_json()

  """


  schema = None
  """Avro schema."""

  def __init__(self, client, hdfs_path, parts=None):
    self._client = client
    self._parts = client.parts(hdfs_path, parts)
    if not self._parts:
      raise HdfsError('No Avro file found at %r.', hdfs_path)

    def _reader():
      """Record generator over all part-files."""
      hdfs_reader = None
      paths = sorted(self._parts)
      try:
        for index, path in enumerate(paths):
          hdfs_reader = client.read(path)
          with _DataFileReader(hdfs_reader) as reader:
            if not index:
              yield reader.schema
              # avoids having to expose the readers directly
            for record in reader:
              yield record
      except GeneratorExit:
        if hdfs_reader:
          hdfs_reader.close()
          # close connection if the reader gets killed early

    self.records = _reader()
    self.schema = self.records.next()

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    self.records.close()

  def __iter__(self):
    return self.records

  @property
  def length(self):
    """Total reader length in bytes."""
    return sum(s['length'] for s in self._parts.values())


class AvroWriter(object):

  """Remote Avro file writer.

  :param client: :class:`hdfs.client.Client` instance.
  :param hdfs_path: Remote path.
  :param overwrite: Overwrite any existing file.
  :param schema: Avro schema. If not passed, will be inferred from the first
    record emitted.

  Usage:

  .. code-block:: python

    with AvroWriter(client, 'data.avro') as writer:
      for elem in elems:
        writer.records.send(elem)

  """

  records = None
  """Avro records coroutine.

  This coroutine can be sent dictionaries. These will then be written to HDFS
  when the coroutine closes. Note that dictionaries passed must conform to the
  writer's schema (if no schema was passed to the constructor, it will be
  inferred from the first dictionary passed).

  """

  def __init__(self, client, hdfs_path, overwrite=False, schema=None):
    if schema and not isinstance(schema, av.schema.Schema):
        raise HdfsError('Invalid schema: %r.', schema)
    self._client = client
    self._schema = schema

    def _writer(_schema=self._schema):
      """Records coroutine."""
      writer = None
      try:
        while True:
          obj = (yield)
          if not writer:
            if not _schema: # no schema implies no writer
              _schema = _get_schema(obj)
              self._schema = _schema
            datum_writer = avi.DatumWriter(_schema)
            buf = BytesIO()
            writer = avd.DataFileWriter(buf, datum_writer, _schema)
          writer.append(obj)
      except GeneratorExit: # we are ready to send the data to HDFS
        if writer:
          writer.flush() # make sure everything has been written to the buffer
          buf.seek(0)
          self._client.write(hdfs_path, buf, overwrite=overwrite)
      finally:
        if writer:
          writer.close()

    self.records = _writer()
    self.records.send(None) # prime coroutine

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    self.records.close()

  @property
  def schema(self):
    """Avro schema."""
    if not self._schema:
      raise HdfsError('Schema not yet inferred.')
    return self._schema


@catch(HdfsError)
def main():
  """Entry point."""
  args = docopt(__doc__)
  client = Client.from_alias(args['--alias'])
  # set up logging
  logger = lg.getLogger('hdfs')
  logger.setLevel(lg.DEBUG)
  handler = Config().get_file_handler('hdfsavro')
  if handler:
    logger.addHandler(handler)
  try:
    parts = args['--parts'] or '0'
    if ',' in parts:
      parts = [int(p) for p in parts.split(',') if p]
    else:
      parts = int(parts)
  except ValueError:
    raise HdfsError('Invalid `--parts` option: %r.', args['--parts'])
  if args['--log']:
    if handler:
      sys.stdout.write('%s\n' % (handler.baseFilename, ))
    else:
      raise HdfsError('No log file active.')
  else:
    avro_file = AvroReader(client, args['PATH'] or '', parts)
    if args['--schema']:
      print dumps(avro_file.schema.to_json(), indent=2)
    elif args['--head']:
      try:
        n_records = int(args['--num'])
      except ValueError:
        raise HdfsError('Invalid `--num` option: %r.', args['--num'])
      for record in islice(avro_file, n_records):
        print dumps(record, indent=2)
    elif args['--sample']:
      num = args['--num']
      frq = args['--freq']
      if frq:
        try:
          freq = float(frq)
        except ValueError:
          raise HdfsError('Invalid `--freq` option: %r.', args['--freq'])
        for record in avro_file:
          if random() <= freq:
            print dumps(record)
      else:
        try:
          n_records = int(num)
        except ValueError:
          raise HdfsError('Invalid `--num` option: %r.', num)
        for record in islice(avro_file, n_records):
          print dumps(record)

if __name__ == '__main__':
  main()



#!/usr/bin/env python
# encoding: utf-8

"""Test Hdfs client interactions with HDFS."""

from hdfs.client import *
from hdfs.util import HdfsError, temppath
from helpers import _TestSession
from nose.tools import eq_, ok_, raises
from requests.exceptions import ConnectTimeout
from shutil import rmtree
from tempfile import mkdtemp
import os
import os.path as osp
import time


def status(response):
  """Helper for requests that return boolean JSON responses."""
  return response.json()['boolean']


class TestLoad(object):

  """Test client loader."""

  def test_bare(self):
    client = Client._from_options(None, {'url': 'foo'})
    ok_(isinstance(client, Client))

  def test_new_type(self):
    class NewClient(Client):
      def __init__(self, url, bar):
        super(NewClient, self).__init__(url)
        self.bar = bar
    client = Client._from_options('NewClient', {'url': 'bar', 'bar': 2})
    eq_(client.bar, 2)

  @raises(HdfsError)
  def test_missing_options(self):
    client = Client._from_options('KerberosClient', {})

  @raises(HdfsError)
  def test_invalid_options(self):
    client = Client._from_options(None, {'foo': 123})

  @raises(HdfsError)
  def test_missing_type(self):
    client = Client._from_options('foo', {})


class TestOptions(_TestSession):

  """Test client options."""

  def test_timeout(self):
    self.client.timeout = 1e-4 # Small enough for it to always timeout.
    try:
      self.client.status('.')
    except ConnectTimeout:
      self.client.timeout = None
    else:
      raise HdfsError('No timeout.')


class TestApi(_TestSession):

  """Test client raw API interactions."""

  def test_list_status_absolute_root(self):
    ok_(self.client._list_status('/'))

  def test_list_status_test_root(self):
    eq_(
      self.client._list_status('').content,
      self.client._list_status(self.client.root).content,
    )

  def test_get_file_status(self):
    status = self.client._get_file_status('').json()['FileStatus']
    eq_(status['type'], 'DIRECTORY')

  def test_get_home_directory(self):
    path = self.client._get_home_directory('').json()['Path']
    ok_('/user/' in path)

  def test_create_file(self):
    path = 'foo'
    self.client._create(path, data='hello')
    ok_(self._file_exists(path))

  def test_delete_file(self):
    path = 'bar'
    self.client._create(path, data='hello')
    ok_(status(self.client._delete(path)))
    ok_(not self._file_exists(path))

  def test_delete_missing_file(self):
    path = 'bar2'
    ok_(not status(self.client._delete(path)))

  def test_rename_file(self):
    paths = ['foo', '%s/bar' % (self.client.root.rstrip('/'), )]
    self.client._create(paths[0], data='hello')
    ok_(status(self.client._rename(paths[0], destination=paths[1])))
    ok_(not self._file_exists(paths[0]))
    eq_(self.client._open(paths[1].rsplit('/', 1)[1]).content, 'hello')
    self.client._delete(paths[1])

  def test_rename_file_to_existing(self):
    paths = ['foo', '%s/bar' % (self.client.root.rstrip('/'), )]
    self.client._create(paths[0], data='hello')
    self.client._create(paths[1], data='hi')
    try:
      ok_(not status(self.client._rename(paths[0], destination=paths[1])))
    finally:
      self.client._delete(paths[0])
      self.client._delete(paths[1])

  def test_open_file(self):
    self.client._create('foo', data='hello')
    eq_(self.client._open('foo').content, 'hello')

  def test_get_file_checksum(self):
    self.client._create('foo', data='hello')
    data = self.client._get_file_checksum('foo').json()['FileChecksum']
    eq_(sorted(data), ['algorithm', 'bytes', 'length'])
    ok_(int(data['length']))

  @raises(HdfsError)
  def test_get_file_checksum_on_folder(self):
    self.client._get_file_checksum('')


class TestResolve(_TestSession):

  def test_resolve_filename(self):
    path = 'fo&o/a?%a'
    encoded = self.client.resolve(path)
    eq_(encoded.split('/')[-2:], ['fo%26o', 'a%3F%25a'])

  def test_create_file_with_reserved_characters(self):
    path = 'fo&o/a?a'
    self.client.write(path, data='hello')
    eq_(''.join(self.client.read(path)), 'hello')

  def test_create_file_with_percent(self):
    # `%` (`0x25`) is a special case because it seems to cause errors (even
    # though the action still goes through). Typical error message will be
    # `"Unknown exception in doAs"`.
    path = 'fo&o/a%a'
    try:
      self.client.write(path, data='hello')
    except HdfsError:
      pass
    eq_(''.join(self.client.read(path)), 'hello')


class TestWrite(_TestSession):

  def test_create_from_string(self):
    self.client.write('up', 'hello, world!')
    self._check_content('up', 'hello, world!')

  def test_create_from_generator(self):
    data = (e for e in ['hello, ', 'world!'])
    self.client.write('up', data)
    self._check_content('up', 'hello, world!')

  def test_create_from_file_object(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('hello, world!')
      with open(tpath) as reader:
        self.client.write('up', reader)
    self._check_content('up', 'hello, world!')

  def test_create_set_permissions(self):
    pass # TODO

  @raises(HdfsError)
  def test_create_to_existing_file_without_overwrite(self):
    self.client.write('up', 'hello, world!')
    self.client.write('up', 'hello again, world!')

  def test_create_and_overwrite_file(self):
    self.client.write('up', 'hello, world!')
    self.client.write('up', 'hello again, world!', overwrite=True)
    self._check_content('up', 'hello again, world!')

  @raises(HdfsError)
  def test_create_and_overwrite_directory(self):
    # can't overwrite a directory with a file
    self.client._mkdirs('up')
    self.client.write('up', 'hello, world!')

  @raises(HdfsError)
  def test_create_invalid_path(self):
    # conversely, can't overwrite a file with a directory
    self.client.write('up', 'hello, world!')
    self.client.write('up/up', 'hello again, world!')


class TestAppend(_TestSession):

  @classmethod
  def setup_class(cls):
    super(TestAppend, cls).setup_class()
    if cls.client:
      try:
        cls.client.write('ap', '')
        # can't append to an empty file
        cls.client.append('ap', '')
        # try a simple append
      except HdfsError as err:
        if 'Append is not supported.' in err.message:
          cls.client = None
          # skip these tests if HDFS isn't configured to support appends
        else:
          raise err

  def test_simple(self):
    self.client.write('ap', 'hello,')
    self.client.append('ap', ' world!')
    self._check_content('ap', 'hello, world!')

  @raises(HdfsError)
  def test_missing_file(self):
    self.client.append('ap', 'hello,')


class TestUpload(_TestSession):

  def test_upload_file(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('hello, world!')
      self.client.upload('up', tpath)
    self._check_content('up', 'hello, world!')

  @raises(HdfsError)
  def test_upload_directory(self):
    tdpath = mkdtemp()
    try:
      self.client.upload('up', tdpath)
    finally:
      os.rmdir(tdpath)

  def test_upload_overwrite(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('hello')
      self.client.upload('up', tpath)
      first_mtime = self._get_mtime('up')
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('there')
      self.client.upload('up', tpath, overwrite=True)
      second_mtime = self._get_mtime('up')
    self._check_content('up', 'there')
    ok_(second_mtime > first_mtime)

  # TODO: reimplement caching
  # def test_upload_overwrite_cached(self):
  #   with temppath() as tpath:
  #     with open(tpath, 'w') as writer:
  #       writer.write('hello')
  #     self.client.upload('up', tpath)
  #     first_mtime = self._get_mtime('up')
  #     self.client.upload('up', tpath, overwrite=True)
  #     second_mtime = self._get_mtime('up')
  #   ok_(second_mtime == first_mtime)

  @raises(HdfsError)
  def test_upload_overwrite_error(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('here')
      self.client.upload('up', tpath)
      self.client.upload('up', tpath)

  def _get_mtime(self, hdfs_path):
    return self.client.status(hdfs_path)['modificationTime']


class TestDelete(_TestSession):

  def test_delete_file(self):
    self.client.write('foo', 'hello, world!')
    self.client.delete('foo')

  def test_delete_empty_directory(self):
    self.client._mkdirs('foo')
    self.client.delete('foo')

  @raises(HdfsError)
  def test_delete_missing_file(self):
    self.client.delete('foo')

  def test_delete_non_empty_directory(self):
    self.client.write('de/foo', 'hello, world!')
    self.client.delete('de', recursive=True)

  @raises(HdfsError)
  def test_delete_non_empty_directory_without_recursive(self):
    self.client.write('de/foo', 'hello, world!')
    self.client.delete('de')


class TestRead(_TestSession):

  def test_read_file(self):
    self.client.write('foo', 'hello, world!')
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        self._read(writer, 'foo')
      with open(tpath) as reader:
        eq_(reader.read(), 'hello, world!')

  @raises(HdfsError)
  def test_read_directory(self):
    self.client._mkdirs('foo')
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        self._read(writer, 'foo')

  @raises(HdfsError)
  def test_read_missing_file(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        self._read(writer, 'foo')

  def test_read_file_from_offset(self):
    self.client.write('foo', 'hello, world!')
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        self._read(writer, 'foo', offset=7)
      with open(tpath) as reader:
        eq_(reader.read(), 'world!')

  def test_read_file_from_offset_with_limit(self):
    self.client.write('foo', 'hello, world!')
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        self._read(writer, 'foo', offset=7, length=5)
      with open(tpath) as reader:
        eq_(reader.read(), 'world')

  def _read(self, writer, *args, **kwargs):
    for chunk in self.client.read(*args, **kwargs):
      writer.write(chunk)


class TestRename(_TestSession):

  def test_rename_file(self):
    self.client.write('foo', 'hello, world!')
    self.client.rename('foo', 'bar')
    self._check_content('bar', 'hello, world!')

  @raises(HdfsError)
  def test_rename_missing_file(self):
    self.client.rename('foo', 'bar')

  @raises(HdfsError)
  def test_rename_file_to_existing_file(self):
    self.client.write('foo', 'hello, world!')
    self.client.write('bar', 'hello again, world!')
    self.client.rename('foo', 'bar')

  def test_rename_file_into_existing_directory(self):
    self.client.write('foo', 'hello, world!')
    self.client._mkdirs('bar')
    self.client.rename('foo', 'bar')
    self._check_content('bar/foo', 'hello, world!')


class TestDownload(_TestSession):

  def setup(self):
    super(TestDownload, self).setup()
    self.parts = {
      'part-r-00000': 'fee',
      'part-r-00001': 'faa',
      'part-r-00002': 'foo',
    }

  @raises(HdfsError)
  def test_missing_dir(self):
    self.client.write('dl', 'hello')
    with temppath() as tpath:
      self.client.download('dl', osp.join(tpath, 'foo'))

  def test_normal_file(self):
    self.client.write('dl', 'hello')
    with temppath() as tpath:
      fname = self.client.download('dl', tpath)
      with open(fname) as reader:
        eq_(reader.read(), 'hello')

  def test_nonpartitioned_file(self):
    partname = 'part-r-00000'
    self.client.write('dl/' + partname, 'world')
    with temppath() as tpath:
      fname = self.client.download('dl/' + partname, tpath)
      with open(fname) as reader:
        eq_(reader.read(), 'world')

  def test_singly_partitioned_file(self):
    partname = 'part-r-00000'
    self.client.write('dl/' + partname, 'world')
    with temppath() as tpath:
      os.mkdir(tpath)
      fname = self.client.download('dl', tpath)
      with open(osp.join(fname, partname)) as reader:
        eq_(reader.read(), 'world')

  def test_partitioned_file_sync(self):
    for name, content in self.parts.items():
      self.client.write('dl/%s' % (name, ), content)
    with temppath() as tpath:
      self.client.download('dl', tpath, n_threads=0)
      self.check_contents(tpath)

  def test_partitioned_file_async_auto(self):
    for name, content in self.parts.items():
      self.client.write('dl/%s' % (name, ), content)
    with temppath() as tpath:
      self.client.download('dl', tpath, n_threads=-1)
      self.check_contents(tpath)

  def test_partitioned_file_async_manual(self):
    for name, content in self.parts.items():
      self.client.write('dl/%s' % (name, ), content)
    with temppath() as tpath:
      self.client.download('dl', tpath, n_threads=2)
      self.check_contents(tpath)

  def test_overwrite_file(self):
    with temppath() as tpath:
      self.client.write('dl', 'hello')
      self.client.download('dl', tpath)
      self.client.write('dl', 'there', overwrite=True)
      fname = self.client.download('dl', tpath, overwrite=True)
      with open(fname) as reader:
        eq_(reader.read(), 'there')

  @raises(HdfsError)
  def test_overwrite_error(self):
    self.client.write('dl', 'hello')
    with temppath() as tpath:
      self.client.download('dl', tpath)
      self.client.download('dl', tpath)

  @raises(HdfsError)
  def test_partitioned_file_to_existing_file(self):
    self.client.write('dl', 'hello')
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('hey')
      self.client.download('dl', tpath)

  # helpers

  def check_contents(self, local_path):
    local_parts = os.listdir(local_path)
    eq_(set(local_parts), set(self.parts))
    # we have all the parts
    for part in local_parts:
      with open(osp.join(local_path, part)) as reader:
        eq_(reader.read(), self.parts[part])
        # their content is correct


class TestStatus(_TestSession):

  def test_directory(self):
    status = self.client.status('')
    eq_(status['type'], 'DIRECTORY')
    eq_(status['length'], 0)

  def test_file(self):
    self.client.write('foo', 'hello, world!')
    status = self.client.status('foo')
    eq_(status['type'], 'FILE')
    eq_(status['length'], 13)

  @raises(HdfsError)
  def test_missing(self):
    self.client.status('foo')


class TestContent(_TestSession):

  def test_directory(self):
    self.client.write('foo', 'hello, world!')
    content = self.client.content('')
    eq_(content['directoryCount'], 1)
    eq_(content['fileCount'], 1)
    eq_(content['length'], 13)

  def test_file(self):
    self.client.write('foo', 'hello, world!')
    content = self.client.content('foo')
    eq_(content['directoryCount'], 0)
    eq_(content['fileCount'], 1)
    eq_(content['length'], 13)

  @raises(HdfsError)
  def test_missing(self):
    self.client.content('foo')


class TestList(_TestSession):

  @raises(HdfsError)
  def test_file(self):
    self.client.write('foo', 'hello, world!')
    self.client.list('foo')

  @raises(HdfsError)
  def test_missing(self):
    self.client.list('foo')

  def test_empty_dir(self):
    self.client._mkdirs('foo')
    eq_(self.client.list('foo'), [])

  def test_dir(self):
    self.client.write('foo/bar', 'hello, world!')
    statuses = self.client.list('foo')
    eq_(len(statuses), 1)
    status = self.client.status('foo/bar')
    status['pathSuffix'] = 'bar'
    eq_(statuses[0], (osp.join(self.client.root, 'foo', 'bar'), status))


class TestWalk(_TestSession):

  def test_file(self):
    self.client.write('foo', 'hello, world!')
    infos = list(self.client.walk('foo'))
    status = self.client.status('foo')
    eq_(len(infos), 1)
    eq_(infos[0], (osp.join(self.client.root, 'foo'), status))

  def test_file_with_depth(self):
    self.client.write('foo', 'hello, world!')
    infos = list(self.client.walk('foo', depth=48))
    status = self.client.status('foo')
    eq_(len(infos), 1)
    eq_(infos[0], (osp.join(self.client.root, 'foo'), status))

  def test_dir_without_depth(self):
    self.client.write('bar/foo', 'hello, world!')
    infos = list(self.client.walk('bar', depth=0))
    status = self.client.status('bar')
    eq_(len(infos), 1)
    eq_(infos[0], (osp.join(self.client.root, 'bar'), status))

  def test_dir_with_depth(self):
    self.client.write('bar/foo', 'hello, world!')
    self.client.write('bar/baz', 'hello again, world!')
    self.client.write('bar/bax/foo', 'hello yet again, world!')
    infos = list(self.client.walk('bar', depth=1))
    eq_(len(infos), 4)


class TestLatestExpansion(_TestSession):

  def test_resolve_simple(self):
    self.client.write('bar', 'hello, world!')
    self.client.write('foo', 'hello again, world!')
    eq_(self.client.resolve('#LATEST'), osp.join(self.client.root, 'foo'))

  def test_resolve_nested(self):
    self.client.write('baz/bar', 'hello, world!')
    self.client.write('bar/bar', 'hello there, world!')
    self.client.write('bar/foo', 'hello again, world!')
    latest = self.client.resolve('#LATEST/#LATEST')
    eq_(latest, osp.join(self.client.root, 'bar', 'foo'))

  def test_resolve_multiple(self):
    self.client.write('bar/bar', 'hello, world!')
    self.client.write('bar/foo', 'hello again, world!')
    latest = self.client.resolve('#LATEST/#LATEST')
    eq_(latest, osp.join(self.client.root, 'bar', 'foo'))

  def test_resolve_multiple_shortcut(self):
    self.client.write('bar/bar', 'hello, world!')
    self.client.write('bar/foo', 'hello again, world!')
    latest = self.client.resolve('#LATEST{2}')
    eq_(latest, osp.join(self.client.root, 'bar', 'foo'))

  @raises(HdfsError)
  def test_resolve_file(self):
    self.client.write('bar', 'hello, world!')
    self.client.resolve('bar/#LATEST')

  @raises(HdfsError)
  def test_resolve_empty_directory(self):
    self.client._mkdirs('bar')
    self.client.resolve('bar/#LATEST')



#!/usr/bin/env python

"""HdfsCLI: a command line interface for WebHDFS."""

from hdfs import __version__
from os import environ
from setuptools import find_packages, setup

# Allow configuration of the CLI alias. This can be helpful since Hadoop 2
# changed `hadoop fs` to `hdfs` (cf. https://github.com/mtth/hdfs/issues/10).
ENTRY_POINT = environ.get('HDFS_ENTRY_POINT', 'hdfs')

setup(
  name='hdfs',
  version=__version__,
  description=__doc__,
  long_description=open('README.rst').read(),
  author='Matthieu Monsch',
  author_email='monsch@alum.mit.edu',
  url='http://hdfscli.readthedocs.org',
  license='MIT',
  packages=find_packages(),
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
  ],
  install_requires=[
    'docopt',
    'requests>=2.0.1',
  ],
  extras_require={
    'avro': ['avro'],
    'kerberos': ['requests-kerberos'],
    'dataframe': ['numpy', 'pandas>=0.14.1', 'fastavro'],
  },
  entry_points={'console_scripts': [
    '%s = hdfs.__main__:main' % (ENTRY_POINT, ),
    '%savro = hdfs.ext.avro:main' % (ENTRY_POINT, ),
  ]},
)



#!/usr/bin/env python
# encoding: utf-8

"""This extension provides helpers for reading and writing `pandas` dataframes.

Dataframes can be read and written into HDFS files stored in one of two
formats: either Avro or CSV (comma separated values, though other separators
such as TAB also commonly used).

This extension requires the presence of `pandas` and `numpy` libraries.  For
reading Avro files, the `fastavro` library is necessary.  This extension
supports downloading multiple HDFS part files in parallel.

"""

from __future__ import absolute_import
from ..client import Client
from ..util import HdfsError
from itertools import chain
from multiprocessing.pool import ThreadPool
import avro
import gzip
import io
import logging
import math
import numpy as np
import operator
import os
import os.path as osp
import pandas as pd
import posixpath
import shutil
import subprocess
import sys
import tempfile
import time


logger = logging.getLogger(__name__)


def _gzip_compress(uncompressed):
  """Utility function that compresses its input using gzip.

  :param uncompressed: String to compress.

  """

  compressed = io.BytesIO()
  with gzip.GzipFile(fileobj=compressed, mode='w') as f:
    f.write(uncompressed)
  return compressed.getvalue()

def _gzip_decompress(compressed):
  """Utility function that decompresses its gzipped-input.

  :param compressed: Bytes to uncompress.

  """

  # zcat substantially faster than built-in gzip library
  p = subprocess.Popen(
      ['zcat'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
  out, _ = p.communicate(compressed)

  return out

def _convert_dtype(dtype):
  """Utility function to convert `numpy` datatypes to their Avro equivalents.

  """
  if np.issubdtype(dtype, np.floating):
    return 'float'
  elif np.issubdtype(dtype, np.integer) \
      or np.issubdtype(dtype, np.unsignedinteger):
    return 'int'
  elif np.issubdtype(dtype, np.character):
    return 'string'
  elif np.issubdtype(dtype, np.bool_):
    return 'boolean'
  else:
    raise HdfsError('Dont know Avro equivalent of type %r.', dtype)


def read_df(client, hdfs_path, format, use_gzip = False, sep = '\t',
  index_cols = None, n_threads = -1, local_dir = None, overwrite=False):
  """Function to read in pandas `DataFrame` from a remote HDFS file.

  :param client: :class:`hdfs.client.Client` instance.
  :param hdfs_path: Remote path, must be a partitioned file.
  :param format: Indicates format of remote file, currently either `'avro'`
    or `'csv'`.
  :param use_gzip: Whether remote file is gzip-compressed or not.  Only
    available for `'csv'` format.
  :param sep: Separator to use for `'csv'` file format.
  :param index_cols: Which columns of remote file should be made index columns
    of `pandas` dataframe.  If set to `None`, `pandas` will create a row
    number index.
  :param n_threads: Number of threads to use for parallel downloading of
    part-files. A value of `None` or `1` indicates that parallelization won't
    be used; `-1` (default) uses as many threads as there are part-files.
  :param local_dir: Local directory in which to save downloaded files. If
    set to `None`, a temporary directory will be used.  Otherwise, if remote
    files already exist in the local directory and are older than downloaded
    version, they will not be downloaded again.

  E.g.:

  .. code-block:: python

    df = read_df(client, '/tmp/data.tsv', 'csv', n_threads=-1)

  """
  is_temp_dir = False

  remote_status = client.status(hdfs_path)
  if remote_status['type'] == 'FILE':
    raise HdfsError('Remote location %r must be a directory containing ' +
      'part-files', hdfs_path)

  try:
    if local_dir is None:
      local_dir = tempfile.mkdtemp()
      logger.info('local_dir not specified, using %r.', local_dir)

    if format == 'csv':
      logger.info('Loading %r CSV formatted data from %r',
        'compressed' if use_gzip else 'uncompressed',
        hdfs_path)

      if sep is None:
        raise HdfsError('sep parameter must not be None for CSV reading.')

      def _process_function(data_files):
          # Loads downloaded csv files and return a pandas dataframe

          PIG_CSV_HEADER = '.pig_header'
          hdfs_header_file = posixpath.join(hdfs_path, PIG_CSV_HEADER)
          client.download(hdfs_header_file, local_dir, overwrite=overwrite)
          merged_files = io.BytesIO()
          with open(osp.join(local_dir, PIG_CSV_HEADER), 'r') as f:
            merged_files.write(f.read())

          for filename in data_files:
            s = time.time()
            with open(filename, 'rb') as f:
              contents = f.read()
              if use_gzip:
                contents = _gzip_decompress(contents)
              merged_files.write(contents)
            delay = time.time() - s
            logger.info('Loaded %s in %0.3fs', filename, delay)

          merged_files.seek(0)

          logger.info("Loading dataframe")

          df = pd.io.parsers.read_csv(merged_files, sep=sep)
          if index_cols is not None:
            df = df.set_index(index_cols)

          merged_files.close()

          return df

    elif format == 'avro':
      logger.info('Loading Avro formatted data from %r', hdfs_path)
      if use_gzip:
        raise HdfsError('Cannot use gzip compression with Avro format.')

      def _process_function(data_files):
        # Loads downloaded avro files and returns a pandas dataframe

        try:
          import fastavro
        except ImportError:
          raise HdfsError('Need to have fastavro library installed ' +
                          'in order to dataframes from Avro files.')

        open_files = []
        try:
          for f in data_files:
            logger.info("Opening %s", f)
            open_files.append( open(f,'rb') )

          reader_chain = chain(*[fastavro.reader(f) for f in open_files])
          df = pd.DataFrame.from_records(reader_chain, index=index_cols)

        finally:
          for f in open_files:
            f.close()

        return df

    else:
      raise ValueError('Unknown data format %s' % (format,) )


    t = time.time()

    lpath = client.download(
      hdfs_path, local_dir, n_threads=n_threads, overwrite=overwrite
    )

    data_files = [osp.join(lpath, fname) for fname in os.listdir(lpath)]
    data_files = sorted(data_files)
    df = _process_function(data_files)

    logger.info('Done in %0.3f', time.time() - t)

  finally:
    if is_temp_dir:
      shutil.rmtree(local_dir)

  return df


def write_df(df, client, hdfs_path, format, use_gzip = False, sep = '\t',
  overwrite = False, n_parts = 1):
  """Function to write a pandas `DataFrame` to a remote HDFS file.

  :param df: `pandas` dataframe object to write.
  :param client: :class:`hdfs.client.Client` instance.
  :param hdfs_path: Remote path.
  :param format: Indicates format of remote file, currently either `'avro'`
    or `'csv'`.
  :param use_gzip: Whether remote file is gzip-compressed or not.  Only
    available for `'csv'` format.
  :param sep: Separator to use for `'csv'` file format.
  :param overwrite: Whether to overwrite files on HDFS if they exist.
  :param n_parts: Indicates into how many part-files to split the dataframe.

  E.g.:

  .. code-block:: python

    df = pd.DataFrame.from_records(
        [{'A' :  1, 'B' :  2},
         {'A' : 11, 'B' : 23}])
    write_df(df, client, '/tmp/data.tsv', 'csv')

  """
  try:
    import pandas as pd
  except ImportError:
    raise HdfsError('pandas library needed for dataframe extension.')

  # Include index columns in output if they do not seem to be a
  # row-id automatically added by `pandas` (resulting in a single,
  # unnamed index column)
  if len(df.index.names) > 1 or df.index.names[0] is not None:
    df = df.reset_index()

  parts_ext = ''
  if format == 'csv':

    def _process_function(df):
      r = df.to_csv(sep=sep, header=False, index=False)
      if use_gzip:
        r = _gzip_compress(r)
      return r

    def _finish_function(df):
      header = df[0:0].to_csv(sep=sep, header=True, index=False).strip()
      header_hdfs_filename = posixpath.join(hdfs_path, '.pig_header')
      client.write(header_hdfs_filename, header + "\n", overwrite=True)

    if use_gzip:
      parts_ext = '.gz'
    logger.info('Writing CSV formatted data to %s', hdfs_path)

  elif format == 'avro':

    def _process_function(df):
      fields_str = [
          '{"name": "%s", "type": "%s"}' % (fldname, _convert_dtype(dtype))
          for fldname, dtype in zip(df.columns, df.dtypes)]
      schema_str = """\
    { "type": "record",
      "name": "dfrecord",
      "fields": [
      """ + ",\n".join(fields_str) + """
      ]
    }"""
      schema = avro.schema.parse(schema_str)
      # Create a 'record' (datum) writer
      rec_writer  = avro.io.DatumWriter(schema)
      out_buffer  = io.BytesIO()
      # Create a 'data file' (avro file) writer
      avro_writer = avro.datafile.DataFileWriter(
          out_buffer,
          rec_writer,
          writers_schema = schema)
      for r in df.to_dict(orient='records'):
        avro_writer.append(r)
      avro_writer.flush()
      r = out_buffer.getvalue()
      avro_writer.close()
      return r

    def _finish_function(df):
      pass

    if use_gzip:
      raise HdfsError('Cannot use gzip compression with Avro format.')
    parts_ext = '.avro'
    logger.info('Writing Avro formatted data to %r', hdfs_path)

  else:
    raise ValueError('Unkown data format %r.' % (format,) )

  t = time.time()
  already_exists_error = False
  try:
    client.status(hdfs_path)
    if overwrite:
      client.delete(hdfs_path, recursive=True)
    else:
      already_exists_error = True
  except HdfsError:
    pass
  if already_exists_error:
    raise HdfsError('%r already exists.', hdfs_path)
  num_rows = len(df)
  rows_per_part = int(math.ceil(num_rows / float(n_parts)))
  for part_num, start_ndx in enumerate(range(0, num_rows, rows_per_part)):
    end_ndx = min(start_ndx + rows_per_part, num_rows)
    client.write(
        posixpath.join(hdfs_path, ('part-r-%05d' % part_num) + parts_ext),
        _process_function(df.iloc[start_ndx:end_ndx]),
        overwrite=False)
  _finish_function(df)
  logger.info('Done in %0.3f', time.time() - t)

